#!/usr/bin/env python3

import time
from typing import Optional

from pycozo.client import Client
from rich.progress import track

from .types import MigrationModule
from .utils.console import fail, info, warn
from .utils.fn import paired


def add_migration(
    client: Client, id: Optional[str], previous_id: Optional[str], created_at: float
) -> None:
    """Add a migration to the migrations_manager table."""

    client.insert(
        "migrations_manager",
        {
            "created_at": created_at,
            "previous_id": previous_id,
            "id": id,
        },
    )

    time.sleep(0.1)


def apply(
    client: Client,
    migration1: MigrationModule,
    migration2: MigrationModule,
    down: bool = False,
    verbose: bool = False,
) -> None:
    # Up: migration1 -> migration2 (run migration2.up)
    # Down: migration1 -> migration2 (run migration1.down)
    op = migration1.down if down else migration2.up

    op_name = (
        f"{migration1.MIGRATION_ID}::down" if down else f"{migration2.MIGRATION_ID}::up"
    )

    previous_id = migration1.MIGRATION_ID
    id = migration2.MIGRATION_ID
    created_at = migration2.CREATED_AT
    assert id or previous_id, "Can't migrate from None to None"

    verbose and info(f"Running `{op_name}`...")
    op(client)

    add_migration(
        client,
        created_at=created_at,
        id=id,
        previous_id=previous_id,
    )


def apply_migrations(
    client: Client,
    migrations: list[MigrationModule],
    down: bool = False,
    _rolling_back: bool = False,
    verbose: bool = False,
) -> bool:
    currently_applied = []
    need_to_roll_back = False
    migration_error = None

    for from_, to in track(paired(migrations), description="Migrating"):
        try:
            apply(client, from_, to, down=down, verbose=verbose)
            currently_applied.append(from_)
        except Exception as e:
            if _rolling_back:
                # We're already rolling back, so we can't roll back again
                # Just fail
                fail(f"Rollback failed: {e}")

            migration_error = e
            need_to_roll_back = True
            break

    if need_to_roll_back:
        migration_error and warn(f"Error: {migration_error}")

        # Roll back the migrations that were applied
        # in reverse order
        warn("Migration failed. Reverting...")

        apply_migrations(
            client,
            currently_applied[::-1],
            down=(not down),
            _rolling_back=True,
            verbose=verbose,
        )

        return False

    return True



#!/usr/bin/env python3

from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Literal, Optional, Protocol, TypedDict

import pandas as pd
from pycozo.client import Client
from rich.table import Table

from .utils.rich import dict_to_table, pretty_transforms

EngineType = Literal["sqlite", "rocksdb", "http"]


@dataclass
class Migration:
    id: Optional[str] = None
    previous_id: Optional[str] = None
    created_at: Optional[float] = None
    migrated_at: Optional[int] = None

    def to_pretty_dict(self) -> dict[str, Any]:
        data = asdict(self)
        pretty_dict = {k: pretty_transforms[k](v) for k, v in data.items()}

        return pretty_dict

    def to_table(self) -> Table:
        return dict_to_table(self.to_pretty_dict())

    @classmethod
    def from_pandas(cls, df: pd.DataFrame) -> list["Migration"]:
        return [cls(**row) for row in df.to_dict(orient="records")]

    @classmethod
    def get_first(cls, df: pd.DataFrame, default: Any = None) -> "Migration":
        items = cls.from_pandas(df)
        return items[0] if items else default


class MigrationModule(Protocol):
    MIGRATION_ID: Optional[str]
    CREATED_AT: Optional[float]

    @staticmethod
    def up(client: Client) -> None:
        ...

    @staticmethod
    def down(client: Client) -> None:
        ...


@dataclass
class MigrationModuleInfo:
    path: str
    id: str
    created_at: int


class CozoConnectionOptions(TypedDict):
    host: str
    auth: str


@dataclass
class AppContext:
    client: Client
    engine: EngineType
    path: Path
    options: CozoConnectionOptions
    migrations_dir: Path



from rich import print
import typer

from ..history import get_history
from ..schema import schema_exists
from ..utils.console import fail
from ..utils.rich import df_to_table, pretty_transforms

from .main import app


@app.command()
def history(ctx: typer.Context):
    client = ctx.obj.client

    if not schema_exists(client):
        fail("Schema does not exist. Run `init` first.")

    history = get_history(client).transform(pretty_transforms)
    history_table = df_to_table(history)

    print(history_table)



#!/usr/bin/env python3

from pycozo.client import Client

from .utils.fn import doesnt_throw

check_schema_query = """
?[id] := *migrations_manager { id }
:limit 1
"""

create_schema_query = """
:create migrations_manager {
    migrated_at: Validity default [floor(now()), true] =>
    id: String?,
    previous_id: String?,
    created_at: Float?,
}"""


@doesnt_throw
def schema_exists(client: Client) -> bool:
    # Ensure the migrate schema exists
    client.run(check_schema_query)
    return True


def create_schema(client: Client) -> None:
    client.run(create_schema_query)



#!/usr/bin/env python3

from rich.console import Console
import typer


console = Console(stderr=True)
error_console = Console(stderr=True)


def warn(msg: str, console=console) -> None:
    console.print(f"[bold yellow]⚠[/bold yellow] {msg}")


def info(msg: str, console=console) -> None:
    console.print(f"[bold blue]•[/bold blue] {msg}")


def fail(msg: str, console=error_console) -> None:
    console.print(f"[bold red]❌[/bold red] {msg}")
    console.print()

    raise typer.Exit(1)


def success(msg: str, console=console) -> None:
    console.print(f"[bold green]✔[/bold green] {msg}")
    console.print()

    raise typer.Exit(0)



#!/usr/bin/env python3

from datetime import datetime
from functools import wraps
from typing import Callable, Any


def format_ts(ts):
    return datetime.fromtimestamp(int(ts))


def doesnt_throw(fn: Callable[..., Any]) -> Callable[..., bool]:
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            fn(*args, **kwargs)
            return True

        except Exception:
            return False

    return wrapper


def maybe_apply(
    fn: Callable[..., Any],
    value: Any,
    reject_if: Callable[[Any], bool] = lambda x: x is None,
) -> Any:
    if reject_if(value):
        return None

    return fn(value)


def paired(lst):
    return [*zip(lst[:-1], lst[1:])]


class create_obj:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)



from pathlib import Path
from typing import Optional
from typing_extensions import Annotated

from pycozo.client import Client
import typer

from ..types import AppContext, CozoConnectionOptions
from ..utils.client import is_connected
from ..utils.console import fail

app = typer.Typer(pretty_exceptions_show_locals=False)


@app.callback()
def main(
    ctx: typer.Context,
    migrations_dir: Annotated[Path, typer.Option(help="Directory to use")] = Path(
        "./migrations"
    ),
    engine: Annotated[str, typer.Option(help="Engine to use")] = "sqlite",
    path: Annotated[
        Path,
        typer.Option(help="Database file (not applicable for mem or http engines)"),
    ] = Path("./cozo.db"),
    host: Annotated[
        str, typer.Option(help="Host to connect to (http engine only)")
    ] = "http://127.0.0.1:9070",
    auth: Annotated[
        Optional[str], typer.Option(help="Auth header (http engine only)")
    ] = None,
):
    """
    A simple migration tool for Cozo databases.
    """

    options: CozoConnectionOptions = {"host": host, "auth": auth}
    client = Client(engine, str(path), options=options)

    if not is_connected(client):
        fail("Could not connect to database")

    migrations_dir.mkdir(parents=True, exist_ok=True)

    app_context: AppContext = AppContext(
        client=client,
        engine=engine,
        path=path,
        options=options,
        migrations_dir=migrations_dir,
    )

    ctx.obj = app_context



from datetime import datetime
import re

import typer

from ..files import get_migration_files
from ..template import migration_template
from ..utils.console import fail, success

from .main import app


@app.command()
def create(id: str, ctx: typer.Context):
    migrations_dir = ctx.obj.migrations_dir

    # id can only contain letters, numbers, and underscores
    # can only start with a lowercase letter
    regex = r"^[a-z][a-zA-Z0-9_]*$"
    if not re.match(regex, id):
        fail("`id` can only contain letters, numbers, and underscores")

    # Check that a migration with same id doesn't already exist
    migration_files = get_migration_files(migrations_dir)
    any_existing = any(m.id == id for m in migration_files)
    if any_existing:
        fail(f"A migration with id `{id}` already exists.")

    # Create the migration file
    ts = datetime.utcnow().timestamp()
    path = migrations_dir / f"migrate_{int(ts)}_{id}.py"

    with open(path, "w") as f:
        f.write(migration_template.format(migration_id=id, created_at=ts))

    success(f"Created migration at: {path}")



#!/usr/bin/env python3

import importlib.util
from pathlib import Path
from types import ModuleType
from typing import cast, Optional

from pycozo.client import Client

from .types import MigrationModule, MigrationModuleInfo
from .utils.fn import create_obj


class NullMigrationModule:
    MIGRATION_ID: Optional[str] = None
    CREATED_AT: Optional[float] = None

    @staticmethod
    def up(client: Client) -> None:
        ...

    @staticmethod
    def down(client: Client) -> None:
        ...


def sanity_check(module: MigrationModule) -> None:
    for attr in ["up", "down", "MIGRATION_ID", "CREATED_AT"]:
        assert hasattr(
            module, attr
        ), f"Missing {attr} attribute in migration:{module.__file__}"


def import_migration_file(path: str | Path) -> MigrationModule:
    path = Path(path)
    module_name: str = path.stem
    spec = importlib.util.spec_from_file_location(module_name, str(path))
    module: ModuleType = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(module)
    sanity_check(module)

    return cast(MigrationModule, module)


def get_adjacent_migrations(
    migration_files: list[MigrationModuleInfo],
    starting_id: Optional[str],
    num: int,
    down: bool = False,
) -> list[MigrationModule]:
    """Get the next num migrations after the starting_id."""

    # Add None to the beginning of the list
    migration_files = [create_obj(id=None)] + migration_files

    # Reverse the list if we're going down
    migration_files = migration_files[::-1] if down else migration_files

    # Get the index of the starting_id
    start = next(
        (i for i, m in enumerate(migration_files) if m.id == starting_id), None
    )

    assert start is not None, f"Migration {starting_id} not found."

    end = start + num + 1

    # Load the migration files
    migrations: list[MigrationModule] = [
        import_migration_file(m_file.path)
        if m_file.id is not None
        else NullMigrationModule()
        for m_file in migration_files[start:end]
    ]

    return migrations



#!/usr/bin/env python3

import os
from pathlib import Path

from .types import MigrationModuleInfo


def get_migration_files(migrations_dir: str | Path) -> list[MigrationModuleInfo]:
    files: list[MigrationModuleInfo] = []

    for f in os.listdir(migrations_dir):
        path = os.path.join(migrations_dir, f)
        if not (os.path.isfile(path) and f.startswith("migrate_")):
            continue

        slug = f.replace("migrate_", "").replace(".py", "")
        timestamp, *id_parts = slug.split("_")
        timestamp = int(timestamp)
        id = "_".join(id_parts)

        files.append(
            MigrationModuleInfo(
                path=path,
                id=id,
                created_at=timestamp,
            )
        )

    return sorted(files, key=lambda f: f.created_at)



from typing import Optional
from typing_extensions import Annotated

import typer

from ..files import get_migration_files
from ..history import get_latest_migration
from ..modules import get_adjacent_migrations
from ..ops import apply_migrations
from ..schema import schema_exists
from ..types import Migration, MigrationModule, MigrationModuleInfo
from ..utils.console import fail, success, info

from .main import app


# TODO: Add a flag to apply all migrations
@app.command()
def apply(
    ctx: typer.Context,
    steps: Annotated[int, typer.Argument()] = 1,
    down: bool = False,
    verbose: bool = False,
):
    client = ctx.obj.client
    migrations_dir = ctx.obj.migrations_dir

    if not schema_exists(client):
        fail("Schema does not exist. Run `init` first.")

    # Get the current migration
    current_migration: Optional[Migration] = get_latest_migration(client)
    current_id: Optional[str] = current_migration.id if current_migration else None

    if down and current_id is None:
        fail("No migrations to rollback.")

    # Find migrations in migrations_dir
    migration_files: list[MigrationModuleInfo] = get_migration_files(migrations_dir)
    if len(migration_files) == 0:
        fail("No migrations found.")

    # Get the next steps migrations from starting_id (can be negative)
    migrations: list[MigrationModule] = get_adjacent_migrations(
        migration_files, current_id, steps, down
    )

    if len(migrations) == 0:
        fail("No migrations to apply.")

    # Print the migration path
    joiner = "[yellow] → [/yellow]"
    migration_ids = [m.MIGRATION_ID or "[NONE]" for m in migrations]

    info(f"Migrate path: {joiner.join(migration_ids)}")

    # Confirm
    confirm = typer.confirm("Are you sure you want to apply these migrations?")
    if not confirm:
        raise typer.Abort()

    info("Migrating the database...")
    applied_successfully = apply_migrations(
        client, migrations, down=down, verbose=verbose
    )

    if applied_successfully:
        success("Database migrated.")
    else:
        fail("Migration failed.")



#!/usr/bin/env python3

from typing import Optional

from pycozo.client import Client

from .types import Migration

get_first_query = """
?[
    migrated_at,
    id,
    previous_id,
    created_at,
] := *migrations_manager {
        migrated_at: validity,
        id,
        created_at,
        previous_id,
    },
    migrated_at = to_int(validity),
    previous_id = null
"""

get_latest_query = """
?[
    migrated_at,
    id,
    previous_id,
    created_at,
] := *migrations_manager {
        migrated_at: validity,
        id,
        created_at,
        previous_id,
        @ "NOW"
    }, migrated_at = to_int(validity)
"""

get_history_query = """
?[
    migrated_at,
    id,
    previous_id,
    created_at,
] := *migrations_manager {
        migrated_at: validity,
        id,
        created_at,
        previous_id,
    }, migrated_at = to_int(validity)

:sort migrated_at
"""

get_migration_path_query = """
edges[
    previous_id,
    id,
] := *migrations_manager {
        id,
        previous_id,
    }

starting[id] <- [[null]]
latest[id] := *migrations_manager { id @ "NOW" }

?[start, goal, path] <~ ShortestPathBFS(
    edges[from, to],
    starting[start_idx],
    latest[goal_idx],
)
"""


def get_history(client: Client) -> list[Migration]:
    results = client.run(get_history_query)
    return results


def get_migration_path(client: Client) -> list[str]:
    results = client.run(get_migration_path_query)
    assert len(results) <= 1, "More than one path returned"

    if len(results) == 0:
        return []

    path = results["path"][0]
    return path


def get_latest_migration(client: Client) -> Optional[Migration]:
    migration = Migration.get_first(client.run(get_latest_query))
    return migration


def get_first_migration(client: Client) -> Optional[Migration]:
    migration = Migration.get_first(client.run(get_first_query))
    return migration



#!/usr/bin/env python3

from typing import Optional

import numpy as np
import pandas as pd
from rich import box
from rich.table import Table

from .fn import maybe_apply, format_ts


def reject_nullish(x):
    return x is None or np.isnan(x)

pretty_transforms = dict(
    created_at=lambda x: maybe_apply(format_ts, x, reject_if=reject_nullish) or "---",
    migrated_at=lambda x: maybe_apply(format_ts, x, reject_if=reject_nullish) or "---",
    id=lambda x: f"[bold green]{x}[/bold green]" if x else "---",
    previous_id=lambda x: f"[bold red]{x}[/bold red]" if x else "---",
)


def dict_to_table(data: dict) -> Table:
    """Convert a dict into a rich.Table obj.
    Args:
        data (dict): A dictionary to be converted to a rich Table.
    Returns:
        Table: a rich Table instance populated with the dictionary values."""

    rich_table: Table = Table(box=box.SQUARE)

    columns, values = list(zip(*data.items()))
    for column in columns:
        rich_table.add_column(str(column))

    rich_table.add_row(*map(str, values))

    return rich_table


def df_to_table(
    pandas_dataframe: pd.DataFrame,
    show_index: bool = True,
    index_name: Optional[str] = None,
) -> Table:
    """Convert a pandas.DataFrame obj into a rich.Table obj.
    Args:
        pandas_dataframe (DataFrame): A Pandas DataFrame to be converted to a rich Table.
        show_index (bool): Add a column with a row count to the table. Defaults to True.
        index_name (str, optional): The column name to give to the index column. Defaults to None, showing no value.
    Returns:
        Table: a rich Table instance populated with the DataFrame values."""

    rich_table: Table = Table(box=box.SQUARE)

    if show_index:
        index_name = str(index_name) if index_name else ""
        rich_table.add_column(index_name)

    for column in pandas_dataframe.columns:
        rich_table.add_column(str(column))

    for index, value_list in enumerate(pandas_dataframe.values.tolist()):
        row = [str(index)] if show_index else []
        row += [str(x) for x in value_list]
        rich_table.add_row(*row)

    return rich_table



from rich import print
import typer

from ..history import get_latest_migration
from ..schema import schema_exists
from ..types import Migration
from ..utils.console import fail

from .main import app


@app.command()
def status(ctx: typer.Context):
    client = ctx.obj.client

    if not schema_exists(client):
        fail("Schema does not exist. Run `init` first.")

    last_migration: Migration = get_latest_migration(client)
    if not last_migration:
        fail("No migrations have been applied yet.")

    print(last_migration.to_table())



import typer

from ..schema import schema_exists, create_schema
from ..utils.console import fail, success, info

from .main import app


@app.command()
def init(ctx: typer.Context):
    client = ctx.obj.client

    if schema_exists(client):
        fail("Schema already exists.")

    else:
        info("Initializing the database")
        create_schema(client)
        success("Schema created.")
